package com.sgbarlow.sourcecontrol;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * This is a class that implements source control using Perforce.
 * It is a top-level class until we figure out which package it's supposed
 * to be in.
 * 
 * Version 0.7 update contributed by gbrail@bea.com
 *
 * @author     sgbarlow
 * 
 * @created    14 May 2003
 */

public class PerforceImpl implements SourceControlInterface {
	private final static String P4_COMMAND = "p4";
	
	
	/**
	 *Constructor for the PerforceImpl object
	 */
	public PerforceImpl() { }
	
	
	/**
	 * Read an input stream and return a string.
	 *
	 * @param  stream           Description of the Parameter
	 * @return                  Description of the Return Value
	 * @exception  IOException  Description of the Exception
	 */
	private String readFromInput(InputStream stream)
	throws IOException {
		InputStreamReader reader =
			new InputStreamReader(new BufferedInputStream(stream),
			"UTF-8");
		StringWriter writer = new StringWriter();
		char[] buf = new char[8192];
		int charsRead;
		do {
			charsRead = reader.read(buf);
			if (charsRead > 0) {
				writer.write(buf, 0, charsRead);
			}
		} while (charsRead >= 0);
		reader.close();
		writer.close();
		
		return new String(writer.getBuffer());
	}
	
	/**
	 * This is an internal method that just execs a process and
	 * returns a string containing the output.
	 */
	private String execProgram(String[] args)
	throws Exception {
		Process proc = Runtime.getRuntime().exec(args);
		String resultText = readFromInput(proc.getInputStream());
		
		int result = proc.waitFor();
		if (result != 0) {
			throw new Exception("P4 exited with result code " + result);
		}
		
		return resultText;
	}
	
	
	/**
	 * This is an internal method that runs a P4 command, waits for the
	 * result, and throws an exception if it fails. It parses the output
	 * from P4 and throws an exception if it indicates an error.
	 *
	 * @param  args           arguments to the P4 command
	 * @exception  Exception  Thrown upon failure of the P4 command
	 */
	private void runP4Command(String[] args)
	throws Exception {
		String[] realArgs = new String[args.length + 2];
		realArgs[0] = P4_COMMAND;
		realArgs[1] = "-s";
		System.arraycopy(args, 0, realArgs, 2, args.length);
		
		String output = execProgram(realArgs);
		
		// Find the output lines that start with "error:" and discard the rest
		String errorMsg = getErrorText(output);
		
		if (errorMsg.length() > 0) {
			throw new Exception("P4 error: " + errorMsg);
		}
	}
	
	/**
	 * Given the result of a P4 command, put together all the "error:" lines
	 * into a single error message.
	 */
	private String getErrorText(String text)
	{
		StringBuffer buf = new StringBuffer();
		StringTokenizer tok = new StringTokenizer(text, ":\r\n");
		while (tok.hasMoreTokens()) {
			if (tok.nextToken(":\r\n").equalsIgnoreCase("error")) {
				if (tok.hasMoreTokens()) {
					buf.append(tok.nextToken("\r\n"));
					buf.append('\n');
				}
			} else {
				tok.nextToken("\r\n");
			}
		}
		return buf.toString();
	}
	
	
	/**
	 *  Description of the Method
	 */
	public void init() {
		// Nothing to do
	}
	
	
	/**
	 * Check out the file from P4.
	 *
	 * @param  file           Description of the Parameter
	 * @param  comment        Description of the Parameter
	 * @exception  Exception  Description of the Exception
	 */
	public void checkoutFile(String file, String comment)
	throws Exception {
		String[] args = new String[2];
		args[0] = "edit";
		args[1] = file;
		runP4Command(args);
	}
	
	
	/**
	 *  Description of the Method
	 *
	 * @param  file           Description of the Parameter
	 * @param  comment        Description of the Parameter
	 * @exception  Exception  Description of the Exception
	 */
	public void checkinFile(String file, String comment)
	throws Exception {
		String[] args = new String[2];
		args[0] = "add";
		args[1] = file;
		runP4Command(args);
	}
	
	
	/**
	 *  Description of the Method
	 *
	 * @param  file           Description of the Parameter
	 * @exception  Exception  Description of the Exception
	 */
	public void undoCheckoutFile(String file)
	throws Exception {
		String[] args = new String[2];
		args[0] = "revert";
		args[1] = file;
		runP4Command(args);
	}
	
	/**
	 * Given a single line of P4 output from "p4 filelog", return a 
	 * FileVersion object.
	 */
	private FileVersion parseVersion(String line)
	throws Exception
	{
		FileVersion version = new FileVersion();
		FileLabel label = new FileLabel();
		version.labels.add(label);
		
		// The format is as follows:
		// info1: #<version> change #<changenum> <action> on <date> <time> by 
		//   <user> (<type>) '<comment>'
		// We will simply parse the file, tokenized by white space.
		// If we fail, we'll throw an exception and that'll be it
		StringTokenizer tok = new StringTokenizer(line, " \t");
		
		try {
			tok.nextToken();
			version.version = tok.nextToken();
			if (version.version.charAt(0) == '#') {
				version.version = version.version.substring(1);
			}
			tok.nextToken();
			label.label = tok.nextToken();
			version.action = tok.nextToken();
			tok.nextToken();
			version.date = label.date = tok.nextToken();
			version.time = label.time = tok.nextToken();
			tok.nextToken();
			version.user = label.user = tok.nextToken();
			tok.nextToken();
			version.comment = tok.nextToken();
		} catch (NoSuchElementException nsee) {
			throw new Exception(
			"Perforce did not return the results in the expected format");
		}
		
		return version;
	}
	
	
	public Vector historyFile(String file)
	throws Exception
	{
		String[] args = new String[5];
		args[0] = P4_COMMAND;
		args[1] = "-s";
		args[2] = "filelog";
		args[3] = "-t";
		args[4] = file;
		
		String text = execProgram(args);
		Vector ret = new Vector();
		
		// Parse each line of the output, looking for errors and actual history
		StringTokenizer lineTok = new StringTokenizer(text, "\r\n");
		while (lineTok.hasMoreTokens()) {
			String line = lineTok.nextToken();
			if (line.startsWith("error:")) {
				throw new Exception(getErrorText(text));
			} else if (line.startsWith("info1:")) {
				FileVersion version = parseVersion(line);
				ret.add(version);
			}
		}
		
		return ret;     
	}
	
	/**
	 *  Description of the Method
	 *
	 * @param  file           Description of the Parameter
	 * @return                Description of the Return Value
	 * @exception  Exception  Description of the Exception
	 */
	public String diffFile(String file)
	throws Exception {
		String[] args = new String[5];
		args[0] = P4_COMMAND;
		args[1] = "-s";
		args[2] = "diff";
		args[3] = "-dc";
		args[4] = file;
		
		String diffText = execProgram(args);
		
		if ((diffText != null) && diffText.startsWith("error:")) {
			throw new Exception(getErrorText(diffText));
		} else {
			return diffText;
		}
	}
	
	
	/**
	 *  Gets the file attribute of the PerforceImpl object
	 *
	 * @param  file           Description of the Parameter
	 * @param  path           Description of the Parameter
	 * @param  version        Description of the Parameter
	 * @exception  Exception  Description of the Exception
	 */
	public void getFile(String file, String path, String version)
	throws Exception {
		// First, calculate the output file name
		File sourceFile = new File(file);
		File destFile = new File(path, sourceFile.getName());
		if (destFile.isFile()) {
			destFile.delete();
		}
		
		// Now run the command
		String[] args = new String[4];
		args[0] = "print";
		args[1] = "-o";
		args[2] = destFile.getCanonicalPath();
		if ((version != null) && (version.length() > 0)) {
			args[3] = file + "#" + version;
		} else {
			args[3] = file;
		}
		runP4Command(args);
	}

	public void addFile(String file, String filePath, String comment)
	throws Exception
	{
		throw new Exception("PerforceImpl - Diff is not supported!");
	}

}


